# $Id: fsdict.py,v 1.3 2004/09/05 09:38:16 phil Exp $

#from __future__ import generators

import os
import os.path
import sys
java = sys.platform.startswith('java')

verbose_locking = 0

class LockError(Exception):
    pass

try:
    from _munge import munge
except ImportError:
    print "fsdict: can't find _munge; using pure Python version of 'munge' (slower)"
    def munge(key):
        s = []
        add = s.append
        for c in key:
            if c.isalnum():
                add(c)
            else:
                add('_%02x' % ord(c))
        return ''.join(s)

    def dehex(hexstring):
        n = 0
        #print "dehexing",hexstring
        for c in hexstring.lower():
            #print "c",c,
            n <<= 4
            asc = ord(c)
            #print asc,
            if asc < 65:
                #print 'digit',asc-48,
                n |= asc - 48 # ord('0') == 48
            else:
                #print "char",asc-97+10
                n |= asc - 97 + 10 # ord('a') == 97
            #print "->",n
        #print "that makes",n
        return n
    def unmunge(munged):
        key = []
        i = 0
        max = len(munged)
        while i < max:
            c = munged[i]
            if c == '_':
                key.append(chr(dehex(munged[i+1:i+3])))
                i += 2
            else:
                key.append(c)
            i += 1
        #print "key",key
        return ''.join(key)

class fsdict:
    def __init__(self, path, parent=None, use_locks=1):
        self.use_locks = use_locks
        if not os.path.exists(path):
            os.mkdir(path)
        self.path = path
        if parent is not None:
            self.basepath = parent.basepath
            self.parent = parent
        else:
            self.basepath = path
            self.parent = None
        self.locked = 0
    def __del__(self):
        if self.locked:
            if verbose_locking: print "object in path %s deleted while locked; unlocking ..." % self.path
            self.unlock()
    def __kpath(self, key, islock=0):
        if not islock:
            assert not key.startswith('.')
            key = munge(key)
        return os.path.join(self.path, key)
    def __subdict(self, path):
        return fsdict(path, self)
    def __ensurelocked(self):
        if not self.use_locks: return
        o = self
        while o is not None:
            if o.locked: return
            o = o.parent
        raise LockError, "Can't do this if not locked"
    def __getitem__(self, key):
        kp = self.__kpath(key)
        if not os.path.exists(kp):
            raise KeyError, "Key %s (path %s) not found" % (`key`, kp)
        if os.path.isdir(kp):
            return self.__subdict(kp)
        else:
            f = open(kp)
            try:
                s = f.read()
            finally:
                f.close()
            if java:
                return s
            else:
                return s.decode('utf-8')
    def __setitem__(self, key, val):
        self.__ensurelocked()
        kp = self.__kpath(key)
        if type(val) == type({}):
            subf = self.__subdict(kp)
            for k,v in val.items():
                subf[k] = v
        else:
            if not java:
                val = val.encode('utf-8')
            f = open(kp, 'wt')
            try:
                f.write(val)
            finally:
                f.close()
    def appendtoitem(self, key, val):
        self.__ensurelocked()
        kp = self.__kpath(key)
        assert type(val) != type({}), "can't append text to a dict!"
        if not java:
            val = val.encode('utf-8')
        f = open(kp, 'at')
        try:
            f.write(val)
        finally:
            f.close()
    def __delitem__(self, key):
        self.__ensurelocked()
        kp = self.__kpath(key)
        if os.path.isdir(kp):
            subf = self.__subdict(kp)
            for k in subf.keys():
                del subf[k]
            os.rmdir(kp)
        else:
            os.remove(kp)
    def __lockpath(self):
        return self.__kpath('.lock', islock=1)
    def lock(self):
        # lock self, then go up the chain and make sure there aren't
        # any other locks active ...
        mylp = self.__lockpath()
        if verbose_locking: print "putting a lock at %s" % mylp
        try:
            os.mkdir(mylp)
        except OSError:
            raise LockError, "Can't lock as a lock is already in place"
        try:
            o = self.parent
            while o is not None:
                plock = os.path.join(o.path, '.lock')
                if os.path.isdir(plock):
                    raise LockError, "Can't lock %s as a lock already exists in %s" % (self.path, plock)
                o = o.parent
            self.locked = 1
        except:
            if verbose_locking: print "failed; removing lock at %s" % mylp
            os.rmdir(mylp)
            raise
    def unlock(self):
        assert self.locked
        if verbose_locking: print "unlock:",self.__lockpath()
        os.rmdir(self.__lockpath())
        self.locked = 0
    def get(self, key, failobj):
        if self.has_key(key):
            return self[key]
        return failobj
    def setdefault(self, key, val):
        if not self.has_key(key):
            self[key] = val
        return self[key]
    def keys(self):
        return [unmunge(x) for x in os.listdir(self.path) if not x.startswith('.')]
    def __len__(self):
        l = 0
        for x in os.listdir(self.path):
            if not x.startswith('.'):
                l += 1
        return l
    def items(self):
        return [(k, self[k]) for k in self.keys()]
        #for k in self.keys():
        #    yield (k, self[k])
    def has_key(self, key):
        return os.path.exists(self.__kpath(key))

if __name__ == '__main__':
    verbose_locking = 1
    test_dir = os.path.expanduser('~/.fsdict.test')
    print "1",
    f = fsdict(test_dir)
    try:
        f['asdf'] = 'this shouldn\'t work'
        raise Exception("test failed: shouldn't be able to set an item without a lock")
    except LockError:
        pass
    print "2",
    f.lock()
    f['foo'] = 'bar'
    assert f['foo'] == 'bar'
    print "3",
    f.unlock()
    try:
        f['two'] = "this shouldn't work either"
        raise Exception("test failed: didn't unlock properly")
    except LockError:
        pass
    print "4",
    f.lock()
    f['baz'] = {'boz': 'asdf', 'ooo': 'aaa'}
    assert f['baz']['ooo'] == 'aaa'
    print "5",
    del f['baz']['ooo']
    assert not f['baz'].has_key('ooo')
    print "6",
    del f['baz']
    assert not f.has_key('baz')
    print "7",
    f['sub'] = {}
    x = f['sub']
    weirdkey = '*&!%#@$^%'
    x[weirdkey] = weirdkey
    assert x[weirdkey] == weirdkey, "can't look up a key with a mungeable name"
    assert x.keys() == [weirdkey], "keys() doesn't unmunge key names properly"
    for y in range(100):
        x[`y`] = `y`
    del x
    assert f['sub']['52'] == '52'
    print "8",
    g = fsdict(test_dir)
    s = g['sub']
    try:
        s.lock()
        raise Exception("test failed: shouldn't be able to lock another fsdict on top of an existing one")
    except LockError:
        pass
    print "9",
    f.unlock()
    s.lock()
    assert s['42'] == '42'
    s.unlock()
    print "del s"
    del s
    print "del g"
    del g
    f.lock()
    for k in f.keys():
        del f[k]
    print "del f (will get a warning about unlocking during __del__)"
    del f
    os.rmdir(test_dir)
